/*
    Copyright (C) 2021  HisArmMan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
// helper functions that dont directly deal with the game itself
#include <goodbrew/config.h>
#include <goodbrew/base.h>
#include <goodbrew/graphics.h>
#include <goodbrew/images.h>
#include <goodbrew/sound.h>
#include <goodbrew/text.h>
#include <goodbrew/paths.h>
#include <goodbrew/controls.h>
#include <stdlib.h>
#include <string.h>
#include "assistance.h"
u64 tohdtime(mltime in){
	return (in/(double)1000)*getHDTimeRes();
}
mltime tomltime(u64 in){
	return (in/(double)getHDTimeRes())*1000;
}
int slideonfromlow(int dest, int size, u64 curTime, u64 endTime, mltime deltaTime, char reverse){
	return size*-1+partMoveFillsR(curTime, endTime, deltaTime, dest+size, reverse);
}
int slideonfromhigh(int dest, int maxval, u64 curTime, u64 endTime, mltime deltaTime, char reverse){
	return dest+partMoveFillsR(curTime, endTime, deltaTime, maxval-dest, !reverse);
}
/* double partMoveRepeats(u64 curTime, u64 startTime, mltime singleRotation, double maxValue){ */
/* 	u64 diff = curTime-startTime; */
/* 	return maxValue*((diff%singleRotation)/(double)tohdtime(singleRotation)); */
/* } */
double partMoveFills(u64 curTime, u64 endTime, mltime deltaTime, double maxValue){
	mltime timepassed;
	if (curTime>endTime){
		timepassed=deltaTime+tomltime(curTime-endTime);
	}else{
		timepassed=deltaTime-tomltime(endTime-curTime);
	}
	return (timepassed/(double)deltaTime)*maxValue;
}
double partMoveEmpties(u64 curTime, u64 endTime, mltime deltaTime, double max){
	double r =partMoveFills(curTime, endTime, deltaTime, max);
	if (r>=max){
		return 0;
	}
	return max-r;
}
double partMoveFillsR(u64 curTime, u64 endTime, mltime deltaTime, double maxValue, char reverse){
	if (reverse){
		return partMoveEmpties(curTime, endTime, deltaTime, maxValue);
	}else{
		return partMoveFills(curTime, endTime, deltaTime, maxValue);
	}
}
int wrapnum(int _passed, int _min, int _max){
	if (_passed<_min){
		return _max-(_min-_passed-1);
	}
	if (_passed>_max){
		return _min+(_passed-_max-1);
	}
	return _passed;
}
double limitNumd(double _passed, double _min, double _max){
	if (_passed<_min){
		return _min;
	}if (_passed>_max){
		return _max;
	}
	return _passed;
}
int limitNumdW(double* passed, double min, double max){
	if (*passed<min){
		*passed=min;
		return 1;
	}if (*passed>max){
		*passed=max;
		return 1;
	}
	return 0;
}
int limitNum(int _passed, int _min, int _max){
	if (_passed<_min){
		return _min;
	}if (_passed>_max){
		return _max;
	}
	return _passed;
}
int limitNumW(int* passed, int min, int max){
	if (*passed<min){
		*passed=min;
		return 1;
	}if (*passed>max){
		*passed=max;
		return 1;
	}
	return 0;
}
int ezcenter(int small, int big){
	return (big-small)/2;
}
void drawHallowRect(int _x, int _y, int _w, int _h, int _thick, int _r, int _g, int _b, int _a){
	drawRectangle(_x,_y,_thick,_h,_r,_g,_b,_a);
	drawRectangle(_x+_w-_thick,_y,_thick,_h,_r,_g,_b,_a);
	drawRectangle(_x,_y,_w,_thick,_r,_g,_b,_a);
	drawRectangle(_x,_y+_h-_thick,_w,_thick,_r,_g,_b,_a);
}
void drawborderrect(int x, int y, int w, int h, int thick, unsigned char r, unsigned char g, unsigned char b, unsigned char r2, unsigned char g2, unsigned char b2, unsigned char a){
	drawRectangle(x,y,w,h,r,g,b,a);
	drawRectangle(x+thick,y+thick,w-thick*2,h-thick*2,r2,g2,b2,a);
}
int randbetween(int mininclusive, int maxinclusive){
	if (mininclusive>maxinclusive){
		return maxinclusive;
	}
	return rand()%(maxinclusive-mininclusive+1)+mininclusive;
}
int intmax(int a, int b){
	if (a>b){
		return a;
	}
	return b;
}
int intmin(int a, int b){
	if (a<b){
		return a;
	}
	return b;
}
int getscaledh(int w, crossTexture* i){
	return (w/(double)getTextureWidth(i))*getTextureHeight(i);
}
int getscaledw(int h, crossTexture* i){
	return (h/(double)getTextureHeight(i))*getTextureWidth(i);
}
void drawTextureSizedTintAlphaRotate(crossTexture* tex, float destX, float destY, int destW, int destH, unsigned char r, unsigned char g, unsigned char b, unsigned char a, double angle){
	drawTexturePartSizedTintAlphaRotate(tex, destX, destY, destW, destH, 0, 0, getTextureWidth(tex), getTextureHeight(tex), r, g, b, a, angle, 0);
}
void drawTextureSizedAlphaRotate(crossTexture* tex, float destX, float destY, int destW, int destH, unsigned char a, double angle){
	drawTexturePartSizedTintAlphaRotate(tex, destX, destY, destW, destH, 0, 0, getTextureWidth(tex), getTextureHeight(tex), 255,255,255, a, angle, 0);
}
void fixwh(int* w, int* h, crossTexture* tex){
	if (*w==-1){
		*w=getscaledw(*h,tex);
	}else if (*h==-1){
		*h=getscaledh(*w,tex);
	}
}
void drawTextureSizedAlphaRotateFlip(crossTexture* tex, float destX, float destY, int destW, int destH, unsigned char a, double angle, char flip){
	fixwh(&destW,&destH,tex);
	drawTexturePartSizedTintAlphaRotate(tex, destX, destY, destW, destH, 0, 0, getTextureWidth(tex), getTextureHeight(tex), 255,255,255, a, angle, flip);
}
void drawTextureSizedAlphaRotateFlip00(crossTexture* tex, float destX, float destY, int destW, int destH, unsigned char a, double angle, char flip){
	fixwh(&destW,&destH,tex);
	drawTextureSizedTintAlphaRotate00(tex, destX, destY, destW, destH, 255,255,255, a, angle, flip);
}
void fitinbox(crossTexture* tex, int maxw, int maxh, int* retw, int* reth){
	int w = getTextureWidth(tex);
	int h = getTextureHeight(tex);
	double wscale=maxw/(double)w;
	double hscale=maxh/(double)h;
	double scale=wscale<hscale ? wscale : hscale;
	*retw=w*scale;
	*reth=h*scale;
}
int tomultipleof(int in, int multipleof){
	int rem=in % multipleof;
	if (rem){
		return (in-rem)+multipleof;
	}
	return in;
}
///
struct randrro{ // rand range once
	int min;
	int numleft;
	char* used;
};
void randrrosetused(struct randrro* r, int a){
	r->used[a-r->min]=1;
	r->numleft--;
}
static void initrandrro(struct randrro* target, int minincl, int maxincl){
	target->numleft=maxincl-minincl+1;
	target->min=minincl;
	target->used=calloc(1,sizeof(char)*target->numleft);
}
struct randrro* newrandrangeonce(int minincl, int maxincl){
	struct randrro* ret = malloc(sizeof(struct randrro));
	initrandrro(ret,minincl,maxincl);
	return ret;
}
void freerandrro(struct randrro* r){
	free(r->used);
	free(r);
}
int randrronext(struct randrro* r){
	if (r->numleft==0){
		return r->min-1;
	}
	int next=rand()%r->numleft+1;
	for (int i=0;;++i){
		if (!r->used[i]){
			next--;
			if (next==0){
				r->used[i]=1;
				r->numleft--;
				return i+r->min;
			}
		}
	}
}
//
struct randso{ // rand specific once
	struct randrro p;
	int* nummap;
	int ondoneret;
};
struct randso* newrandspecificonce(int* nummap, int count, int ondoneret){
	struct randso* ret = malloc(sizeof(struct randso));
	initrandrro((void*)ret,0,count-1);
	ret->nummap=nummap;
	ret->ondoneret=ondoneret;
	return ret;
}
void freerandso(struct randso* r){
	freerandrro((void*)r);
}
int randsonext(struct randso* r){
	int i=randrronext((void*)r);
	if (i<0){
		return r->ondoneret;
	}
	return r->nummap[i];
}
///
void sadflash(){
	startDrawing();
	drawRectangle(0,0,getScreenWidth(),getScreenHeight(),255,0,0,255);
	endDrawing();
	wait(500);
}
static void panicerrorscreen(){
	while(1){
		controlsStart();
		if (wasJustPressed(BUTTON_START)){
			exit(1);
		}
		controlsEnd();
		controlsReset();
		startDrawing();
		drawRectangle(0,0,20,100,255,0,0,255);
		drawRectangle(20,0,30,15,255,0,0,255);
		drawRectangle(20,35,30,15,255,0,0,255);
		endDrawing();
	}
}
crossTexture* loadvpkpng(const char* path){
	char* fixedPath = fixPathAlloc(path,TYPE_EMBEDDED);
	crossTexture* tempTex = loadImage(fixedPath);
	if (tempTex==NULL){
		fprintf(stderr,"%s\n",fixedPath);
		panicerrorscreen();
	}
	free(fixedPath);
	return tempTex;
}
crossSE* loadvpkse(const char* path){
	char* fixedPath = fixPathAlloc(path,TYPE_EMBEDDED);
	crossSE* ret = loadSound(fixedPath);
	if (ret==NULL){
		fprintf(stderr,"%s\n",fixedPath);
		panicerrorscreen();
	}
	free(fixedPath);
	return ret;
}
crossMusic* loadvpkmusic(const char* path){
	char* fixedPath = fixPathAlloc(path,TYPE_EMBEDDED);
	crossMusic* ret = loadMusic(fixedPath);
	if (ret==NULL){
		fprintf(stderr,"%s\n",fixedPath);
		panicerrorscreen();
	}
	free(fixedPath);
	return ret;
}
void loadimagerangeto(crossTexture** dest, const char* formatstr, int minNum, int maxNum){
	char buff[strlen(formatstr)+10];
	for (int i=0;i<=(maxNum-minNum);++i){
		sprintf(buff,formatstr,i+minNum);
		dest[i]=loadvpkpng(buff);
	}
}
crossTexture** loadimagerange(const char* formatstr, int minNum, int maxNum){
	crossTexture** ret = malloc(sizeof(crossTexture*)*(maxNum-minNum+1));
	loadimagerangeto(ret, formatstr, minNum, maxNum);
	return ret;
}
void freeimagerange(crossTexture** range, int minNum, int maxNum){
	int numnums=(maxNum-minNum+1);
	for (int i=0;i<numnums;++i){
		freeTexture(range[i]);
	}
	free(range);
}
static int qsortStringComparer(const void* a, const void* b){
	return strcmp(*(char**)a,*(char**)b);
}
char** getfilelist(char* _dirPath, int extraslots, int* retlen){
	crossDir dir=openDirectory(_dirPath);
	if (!dirOpenWorked(dir)){
		return malloc(sizeof(char*)*extraslots);;
	}
	int _maxFiles=intmax(5,extraslots+1);
	char** _foundFiles = malloc(sizeof(char*)*_maxFiles);
	crossDirStorage lastStorage;
	int _nFiles;
	for (_nFiles=0;directoryRead(&dir,&lastStorage)!=0;_nFiles++){
		if (_nFiles+extraslots>=_maxFiles){
			_maxFiles*=2;
			_foundFiles = realloc(_foundFiles,sizeof(char*)*_maxFiles);
		}
		_foundFiles[_nFiles] = strdup(getDirectoryResultName(&lastStorage));
	}
	qsort(_foundFiles,_nFiles,sizeof(char*),qsortStringComparer);
	directoryClose(dir);
	*retlen=_nFiles;
	return _foundFiles;
}
void makedatadir(const char* s){
	char* path=fixPathAlloc(s,TYPE_DATA);
	createDirectory(path);
	free(path);
}
char* dupjustfname(const char* path){
	return strdup(strrchr(path,'/')+1);
}
//////////
static char issecondaryutf8(unsigned char c){
	return (c & 192)==128;
}
static int nextutf8(unsigned char* _stringStart, int i){
	while(1){
		if (!issecondaryutf8(_stringStart[++i])){
			return i;
		}
	}
}
static int prevutf8(unsigned char* _stringStart, int i){
	while(i>0){
		if (!issecondaryutf8(_stringStart[--i])){
			return i;
		}
	}
	return -1;
}
// _workable must be made with malloc. its buffer may be changed
void wrapText(const char* _passedMessage, int* _numLines, char*** _realLines, int _maxWidth, crossFont* normalFont){
	int _cachedStrlen = strlen(_passedMessage);
	if (_cachedStrlen==0){
		*_numLines=0;
		if (_realLines!=NULL){
			*_realLines=NULL;
		}
		return;
	}
	unsigned char* _workable = (unsigned char*)strdup(_passedMessage);
	int _lastNewline = -1; // Index
	int _lastCharLen = _cachedStrlen-prevutf8(_workable,_cachedStrlen);
	int i;
	for (i=0;i<=_cachedStrlen-_lastCharLen;){
		char _isBreakChar = (_workable[i]=='\n' || _workable[i]==' ');
		if (_isBreakChar || i>=_cachedStrlen-_lastCharLen){
			char _didChop=0;
			char _oldChar;
			if (_isBreakChar){
				_didChop=1;
				_oldChar = _workable[i];
				_workable[i]='\0'; // Chop the string for textWidth function
			}
			if (textWidth(normalFont,(char*)&(_workable[_lastNewline+1]))>_maxWidth){ // If at this spot the string is too long for the screen
				// Find last word before we went off screen
				int j;
				for (j=prevutf8(_workable,i);j>_lastNewline;j=prevutf8(_workable,j)){
					if (_workable[j]==' '){
						break;
					}
				}
				if (j==_lastNewline){ // Didn't find a stopping point, the line is likely one giant word
					char _didBreak=0;
					// Force a stopping point
					for (j=prevutf8(_workable,i);j>_lastNewline;j=prevutf8(_workable,j)){
						char _cacheChar = _workable[j];
						_workable[j]='\0';
						char _canSplit = (textWidth(normalFont,(char*)&(_workable[_lastNewline+1]))<=_maxWidth);
						_workable[j]=_cacheChar;
						if (_canSplit){
							// The character we're at right now, that's where the split needs to be because it's acting as the null terminator right now
							char* _newBuffer = malloc(_cachedStrlen+2);
							memcpy(_newBuffer,_workable,j);
							_newBuffer[j]='\0';
							_lastNewline=j;
							memcpy(&(_newBuffer[j+1]),&(_workable[j]),_cachedStrlen-(j)+1); // Should also copy null
							free(_workable);
							_workable = (unsigned char*)_newBuffer;
							// Account for new byte
							_cachedStrlen++;
							i++;
							_didBreak=1;
							break;
						}
					}
					// Odd, no part between last new line and here less than _maxWidth. This should not happen, but just in case, put some code to at least do something.
					if (!_didBreak){
						printf("odd\n");
						_workable[i]='\0';
					}
				}else{ // Normal, found what we're looking for, just chop at the end of the last word
					_workable[j]='\0';
					_lastNewline=j;
				}
				if (_didChop && _lastNewline!=i){ // Fix chop if happened
					_workable[i]=_oldChar;
				}
				// No matter what we did, we'll still need to start our check from the last new line
				// _lastPropSet keeps properties from being overwritten
				i=_lastNewline;
			}else{
				if (_didChop){
					if (_oldChar=='\n'){ // we're okay to put the line break here. also it's already chopped.
						_lastNewline=i;
					}else{ // Fix chop if happened
						_workable[i]=_oldChar;
					}
				}
			}
		}
		i=nextutf8(_workable,i);
	}
	*_numLines=1;
	for (i=0;i<_cachedStrlen;++i){
		if (_workable[i]=='\0'){
			*_numLines=(*_numLines)+1;
		}
	}
	if (_realLines!=NULL){
		*_realLines = malloc(sizeof(char*)*(*_numLines));
		int _nextCopyIndex=0;
		for (i=0;i<*_numLines;++i){
			(*_realLines)[i] = strdup((char*)&(_workable[_nextCopyIndex]));
			_nextCopyIndex += 1+strlen((char*)&(_workable[_nextCopyIndex]));
		}
	}
	free(_workable);
}
