/*
    Copyright (C) 2021  HisArmMan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <stdlib.h>
#include <goodbrew/config.h>
#include <goodbrew/base.h>
#include <goodbrew/graphics.h>
#include <goodbrew/images.h>
#include <goodbrew/sound.h>
#include <goodbrew/text.h>
#include <math.h>
#include "main.h"
#include "assistance.h"
#include "gamemisc.h"
///////////////////////////////////////////////////
// CONTROLS
///////////////////////////////////////////////////
int buttonpressed(struct boundbutton* b){
	for (int i=0;i<b->numbuts;++i){
		if (wasJustPressed(b->buts[i])){
			return 1;
		}
	}
	return 0;
}
int buttondown(struct boundbutton* b){
	for (int i=0;i<b->numbuts;++i){
		if (isDown(b->buts[i])){
			return 1;
		}
	}
	return 0;
}
void initboundbut(struct boundbutton* b){
	b->buts=NULL;
	b->numbuts=0;
}
void initboundbutdef(struct boundbutton* b, crossButton defaultbutton){
	initboundbut(b);
	addboundbut(b,defaultbutton);
}
void addboundbut(struct boundbutton* b, crossButton addme){
	b->numbuts+=1;
	b->buts=realloc(b->buts,sizeof(crossButton)*(b->numbuts));
	b->buts[b->numbuts-1]=addme;
}
void freeboundbut(struct boundbutton* b){
	free(b->buts);
}
///////////////////////////////////////////////////
// STAMP
///////////////////////////////////////////////////
struct stampres* loadstampres(){
	return (void*)loadimagerange("assets/img/border/stamp/s%d.png", 0, 2);
}
void freestampres(struct stampres* a){
	freeimagerange((void*)a, 0, 2);
}
void calcstamp(struct stampprops* ret, int insidew, int insideh, int cornerw){
	int circlew=cornerw*(13/(double)20);
	int rectw=cornerw*(7/(double)20);

	int edgepartw=circlew+rectw;

	ret->cornerw=cornerw;
	ret->__numedgex=ceil((insidew-rectw)/(double)edgepartw)+1;
	ret->__numedgey=ceil((insideh-rectw)/(double)edgepartw)+1;

	ret->insidew=edgepartw*ret->__numedgex+rectw;
	ret->insideh=edgepartw*ret->__numedgey+rectw;
	ret->entirew=ret->insidew+cornerw*2;
	ret->entireh=ret->insideh+cornerw*2;
}
void drawstamp(int x, int y, struct stampprops* p, unsigned char a, struct stampres* stampresuncast){
	crossTexture** res=(void*)stampresuncast;
	int cw=p->cornerw; // all images are sized to be as tall as the corner
	int circlew=getscaledw(cw,res[1]);
	int rectw=getscaledw(cw,res[2]);

	// for what reason do we not copy and paste code?
	int curx=x;
	int cury=y;
	drawTextureSizedAlphaRotateFlip(res[0], curx, cury, cw, cw, a, 0, GOODBREW_NOFLIP);
	curx+=cw;
	drawTextureSizedAlphaRotateFlip(res[2], curx, cury, rectw, cw, a, 0, GOODBREW_NOFLIP);
	curx+=rectw;
	for (int i=0;i<p->__numedgex;++i){
		drawTextureSizedAlphaRotateFlip(res[1], curx, cury, circlew, cw, a, 0, GOODBREW_NOFLIP);
		curx+=circlew;
		drawTextureSizedAlphaRotateFlip(res[2], curx, cury, rectw, cw, a, 0, GOODBREW_NOFLIP);
		curx+=rectw;
	}
	// the corner
	drawTextureSizedAlphaRotateFlip(res[0], curx, cury, p->cornerw, cw, a, 0, GOODBREW_HFLIP);
	cury+=cw;
	drawTextureSizedAlphaRotateFlip00(res[2], curx+cw, cury, rectw, cw, a, M_PI/2, GOODBREW_NOFLIP);
	cury+=rectw;
	// right strip
	for (int i=0;i<p->__numedgey;++i){
		drawTextureSizedAlphaRotateFlip00(res[1], curx+cw, cury, circlew, cw, a, M_PI/2, GOODBREW_NOFLIP);
		cury+=circlew;
		drawTextureSizedAlphaRotateFlip00(res[2], curx+cw, cury, rectw, cw, a, M_PI/2, GOODBREW_NOFLIP);
		cury+=rectw;
	}
	drawTextureSizedAlphaRotateFlip00(res[0], curx, cury, p->cornerw, cw, a, 0, GOODBREW_HFLIP | GOODBREW_VFLIP);

	// draw from bottom left corner to bottom right corner
	curx=x;
	drawTextureSizedAlphaRotateFlip(res[0], curx, cury, p->cornerw, cw, a, 0, GOODBREW_VFLIP);
	curx+=cw;
	drawTextureSizedAlphaRotateFlip00(res[2], curx+rectw, cury+cw, rectw, cw, a, M_PI, GOODBREW_NOFLIP);
	curx+=rectw;
	for (int i=0;i<p->__numedgex;++i){
		drawTextureSizedAlphaRotateFlip00(res[1], curx+circlew, cury+cw, circlew, cw, a, M_PI, GOODBREW_NOFLIP);
		curx+=circlew;
		drawTextureSizedAlphaRotateFlip00(res[2], curx+rectw, cury+cw, rectw, cw, a, M_PI, GOODBREW_NOFLIP);
		curx+=rectw;
	}
	// from the top left corner to the bottom left corner
	curx=x;
	cury=y+cw;
	drawTextureSizedAlphaRotateFlip00(res[2], curx, cury+rectw, rectw, cw, a, M_PI/2*3, GOODBREW_NOFLIP);
	cury+=rectw;
	for (int i=0;i<p->__numedgey;++i){
		drawTextureSizedAlphaRotateFlip00(res[1], curx, cury+circlew, circlew, cw, a, M_PI/2*3, GOODBREW_NOFLIP);
		cury+=circlew;
		drawTextureSizedAlphaRotateFlip00(res[2], curx, cury+rectw, rectw, cw, a, M_PI/2*3, GOODBREW_NOFLIP);
		cury+=rectw;
	}

	drawRectangle(x+p->cornerw,y+p->cornerw,p->insidew,p->insideh,0xf2,0xf2,0xf2,a);
}
///////////////////////////////////////////////////
#define UDDAS_DELAY preferreddasdelay
#define UDDAS_REPEAT preferreddasrepeat
void inituddas(struct uddas* u, char isactuallylr){
	u->dasdir=0;
	u->nextdasshift=0;
	if (isactuallylr){
		u->upbutton=BUTTON_LEFT;
		u->downbutton=BUTTON_RIGHT;
	}else{
		u->upbutton=BUTTON_UP;
		u->downbutton=BUTTON_DOWN;
	}
}
int getuddas(struct uddas* m, u64 fTime){
	if (wasJustPressed(m->downbutton) || wasJustPressed(m->upbutton)){
		m->dasdir=(wasJustPressed(m->downbutton) ? 1 : -1);
		m->nextdasshift=fTime+tohdtime(UDDAS_DELAY);
		return m->dasdir;
	}else if (m->nextdasshift){
		if (fTime>=m->nextdasshift){
			if ((m->dasdir==1 && isDown(m->downbutton)) || (m->dasdir==-1 && isDown(m->upbutton))){
				m->nextdasshift=fTime+tohdtime(UDDAS_REPEAT);
				return m->dasdir;
			}else{
				m->nextdasshift=0;
			}
		}
	}else{
		if (isDown(m->downbutton) || isDown(m->upbutton)){
			m->dasdir=(isDown(m->downbutton) ? 1 : -1);
			m->nextdasshift=fTime+tohdtime(UDDAS_DELAY);
			return m->dasdir;
		}
	}
	return 0;
}
///////////////////////////////////////////////////
// POPUP MENU
///////////////////////////////////////////////////
void initpumenu(struct pumenu* dest){
	dest->starttime=getHDTime();
	dest->choice=0;
	inituddas(&dest->das,0);
}
void drawpumenu(struct pumenu* menu, u64 fTime){
	int optionblockh=mainFontHeight+blockw;
	int reqh=menu->numoptions*optionblockh+(menu->numoptions-1)*blockw;
	int handh=optionblockh*2/(double)3;
	int handw=getscaledw(handh,handimg);
	int y=ezcenter(reqh,getScreenHeight());
	mltime deltatime=tomltime(fTime-menu->starttime);
	char dointroanim=deltatime<menu->introtime;
	for (int i=0;i<menu->numoptions;++i){
		int textw=textWidth(mainfont,menu->options[i]);
		int optionblockw=textw+blockw;
		int x = ezcenter(optionblockw,getScreenWidth());
		if (dointroanim){
			int newstartx=i&1 ? getScreenWidth() : optionblockw*-1;
			x=newstartx+partMoveFills(fTime,menu->starttime+tohdtime(menu->introtime),menu->introtime,x-newstartx);
		}
		if (menu->choice==i){
			drawTextureSized(handimg,x-handw-blockw/3,y+ezcenter(handh,optionblockh),handw,handh);
			//x+=handw+blockw/2;
		}
		drawRectangle(x,y,optionblockw,optionblockh,PUMENUINSIDECOLOR,255);
		drawHallowRect(x,y,optionblockw,optionblockh,blockw*.1,PUMENUINNERBORDER,255);
		drawHallowRect(x,y,optionblockw,optionblockh,PUMENUOUTERBORDERW,PUMENUOUTERBORDER,255);

		gbDrawText(mainfont, x+ezcenter(textw,optionblockw), y+ezcenter(mainFontHeight,optionblockh), menu->options[i], 255,255,255);
		
		y+=optionblockh+blockw;
	}
}
int pumenucontrols(struct pumenu* m, u64 fTime){
	m->choice=wrapnum(m->choice+getuddas(&m->das,fTime),0,m->numoptions-1);
	if (wasJustPressed(BUTTON_RIGHT)){
		m->choice=m->numoptions-1;
	}else if (wasJustPressed(BUTTON_LEFT)){
		m->choice=0;
	}else if (menuselectpressed()){
		if (tomltime(fTime-m->starttime)<m->introtime){
			m->starttime=fTime-tohdtime(m->introtime);
		}else{
			return m->choice;
		}
	}
	return -1;
}
///////////////////////////////////////////////////
// CMENU (column menu)
///////////////////////////////////////////////////
static char* cmenugetvaluestr(struct cmenu* m, int i, char* dest, int destsize){
	if (!m->values[i]){
		return "";
	}
	if (m->props){
		if (m->props[i] & CMENUPROP_DISPLAYINT){
			snprintf(dest, destsize, "%d", *((int*)m->values[i]));
			dest[destsize-1]='\0';
			return dest;
		}else if (m->props[i] & CMENUPROP_DISPLAYDOUBLE){
			snprintf(dest, destsize, "%.2f", *((double*)m->values[i]));
			dest[destsize-1]='\0';
			return dest;
		}else if (m->props[i] & CMENUPROP_DISPLAYBOOL){
			return *((char*)m->values[i]) ? "True" : "False";
		}
	}
	return m->values[i];
}
void begincmenuoutro(struct cmenu* m, u64 fTime){
	if (m->isoutro){
		return;
	}
	if (m->enterAnimationEnd){ // we quit while the intro is going on.
		m->enterAnimationEnd=fTime+tohdtime(CMENUENTERANIMTIME*(1-(m->enterAnimationEnd-fTime)/(double)tohdtime(CMENUENTERANIMTIME)));
	}else{
		m->enterAnimationEnd=fTime+tohdtime(CMENUENTERANIMTIME);
	}
	m->isoutro=1;
}
void drawcmenu(struct cmenu* m, u64 fTime){
	if (m->isoutro && m->enterAnimationEnd==0){
		return;
	}
	char hasvalues=(m->values!=NULL);
	int insidepad=blockw/4;
	int betweenpad=blockw/4;
	int optionh=mainFontHeight+insidepad*2;
	int maxh=getScreenHeight()*.9;
	int maxoptions=maxh/(optionh+betweenpad);

	int numdrawop=m->numoptions;
	int firstdrawop=0;
	if (m->numoptions>maxoptions){
		numdrawop=maxoptions;
		if (m->choice>numdrawop/2){
			firstdrawop=intmin(m->choice-numdrawop/2,m->numoptions-maxoptions);
		}
	}
	int maxdrawop=firstdrawop+numdrawop-1;
	int cury=(getScreenHeight()-maxh)/2+ezcenter(numdrawop*optionh+(numdrawop-1)*betweenpad,maxh);
	
	int labelw=m->maxlabelw+insidepad*2;
	int valuew=m->maxvaluew+insidepad*2;
	
	char valuebuff[15];
	int labelx;
	int valuex;
	if (hasvalues){
		labelx=getScreenWidth()/2-labelw-blockw/2;
		valuex=getScreenWidth()/2+blockw/2;
	}else{
		labelx=ezcenter(labelw,getScreenWidth());
		valuex=0;
	}

	char doenteranim=fTime<m->enterAnimationEnd;
	if (doenteranim){
		labelx=slideonfromlow(labelx, labelw, fTime, m->enterAnimationEnd, CMENUENTERANIMTIME, m->isoutro);
		valuex=slideonfromhigh(valuex, getScreenWidth(), fTime, m->enterAnimationEnd, CMENUENTERANIMTIME, m->isoutro);
	}

	if (firstdrawop!=0){
		int y=cury-mainFontHeight/1.5;
		if (doenteranim){
			y=slideonfromlow(y, mainFontHeight, fTime, m->enterAnimationEnd, CMENUENTERANIMTIME, m->isoutro);
		}
		gbDrawText(mainfont, ezcenter(textWidth(mainfont,"^"),getScreenWidth()), y, "^", 0,0,0);
	}
	
	for (int i=firstdrawop;i<=maxdrawop;++i){
		unsigned char br=17;
		unsigned char bg=0;
		unsigned char bb=95;
		unsigned char mr=42;
		unsigned char mg=0;
		unsigned char mb=234;
		if (i==m->choice){
			br=255;
			bg=0;
			bb=0;
			mr=137;
			mg=0;
			mb=0;
		}
		drawborderrect(labelx, cury, labelw, optionh, insidepad/2, br,bg,bb,mr,mg,mb, 255);
		gbDrawText(mainfont, labelx+insidepad, cury+insidepad, m->labels[i], 255,255,255);

		if (hasvalues && m->values[i]){
			drawborderrect(valuex, cury, valuew, optionh, insidepad/2, br,bg,bb,mr,mg,mb, 255);
			gbDrawText(mainfont, valuex+insidepad, cury+insidepad, cmenugetvaluestr(m, i, valuebuff, sizeof(valuebuff)), 255,255,255);
			if (i==m->choice && (m->props && (m->props[i] & CMENUPROP_LEFTRIGHT))){
				gbDrawText(mainfont, labelx-betweenpad-textWidth(mainfont,"<"), cury+insidepad, "<", 0,0,0);
				gbDrawText(mainfont, valuex+valuew+betweenpad, cury+insidepad, ">", 0,0,0);
			}
		}
		
		cury+=optionh+betweenpad;
	}
	if (maxdrawop!=m->numoptions-1){
		int y=cury-betweenpad/1.5;
		if (doenteranim){
			y=slideonfromhigh(y, getScreenHeight(), fTime, m->enterAnimationEnd, CMENUENTERANIMTIME, m->isoutro);
		}
		gbDrawText(mainfont, ezcenter(textWidth(mainfont,"v"),getScreenWidth()), y, "v", 0,0,0);
	}
}
int _lowcmenucontrols(struct cmenu* m, u64 fTime){
	m->choice=wrapnum(m->choice+getuddas(&m->ddas, fTime),0,m->numoptions-1);
	if (m->props && m->props[m->choice] & CMENUPROP_LEFTRIGHT){
		return getuddas(&m->rdas, fTime);
	}else{
		if (wasJustPressed(BUTTON_A)){
			return 1;
		}
	}
	return 0;
}
int cmenucontrols(struct cmenu* m, u64 fTime){
	if (m->enterAnimationEnd){
		if (fTime>=m->enterAnimationEnd || menuselectpressed()){
			m->enterAnimationEnd=0;
		}
	}
	if (m->isoutro){
		return 0;
	}
	int r=_lowcmenucontrols(m, fTime);
	if (r && fTime<m->enterAnimationEnd){
		m->enterAnimationEnd=fTime;
		return 0;
	}
	return r;
}
void cmenuvaluechange(struct cmenu* m, int i){
	char valuebuff[15];
	m->maxvaluew=intmax(textWidth(mainfont,cmenugetvaluestr(m, i, valuebuff, sizeof(valuebuff))),m->maxvaluew);
}
void cmenulabelchange(struct cmenu* m, int i){
	m->maxlabelw=intmax(textWidth(mainfont,m->labels[i]),m->maxlabelw);
}
void cmenuvaluechangeall(struct cmenu* m){
	if (m->values){
		int valuew=1;
		char valuebuff[15];
		for (int i=0;i<m->numoptions;++i){
			int neww=textWidth(mainfont,cmenugetvaluestr(m, i, valuebuff, sizeof(valuebuff)));
			valuew=intmax(valuew,neww);
		}
		m->maxvaluew=valuew;
	}
}
void cmenulabelchangeall(struct cmenu* m){
	int labelw=1;
	for (int i=0;i<m->numoptions;++i){
		int neww=textWidth(mainfont,m->labels[i]);
		labelw=intmax(labelw,neww);
	}
	m->maxlabelw=labelw;
}
void initcmenu(struct cmenu* m, u64 fTime){
	m->maxvaluew=1;
	cmenulabelchangeall(m);
	cmenuvaluechangeall(m);
	m->enterAnimationEnd=fTime+tohdtime(CMENUENTERANIMTIME);
	m->choice=0;
	m->isoutro=0;
	inituddas(&m->ddas,0);
	inituddas(&m->rdas,1);
}
int cmenuoutrodone(struct cmenu* c, u64 fTime){
	return c->isoutro && fTime>=c->enterAnimationEnd;
}
int simplecmenu(char** labels, int numlabels, int defaultchoice, void(*extradrawfunc)(struct cmenu*,u64)){
	struct cmenu c;
	c.numoptions=numlabels;
	c.labels=labels;
	c.values=NULL;
	c.props=NULL;
	initcmenu(&c,getHDTime());
	c.choice=defaultchoice;
	if (c.choice>=c.numoptions){
		c.choice=c.numoptions-1;
	}
	int didcancel=0;
	while(1){
		u64 fTime=getHDTime();
		controlsStart();
		int didselect=cmenucontrols(&c,fTime);
		if (!c.isoutro){
			if (didselect || wasJustPressed(BUTTON_START) || (wasJustPressed(BUTTON_B) && (didcancel=1))){
				playSound(menuselectsound,0);
				begincmenuoutro(&c, fTime);
			}
		}
		controlsEnd();
		startDrawing();
		drawbackground();
		if (extradrawfunc){
			extradrawfunc(&c,fTime);
		}
		drawcmenu(&c,fTime);
		endDrawing();
		if (cmenuoutrodone(&c, fTime)){
			break;
		}
	}
	if (didcancel){
		return -1;
	}
	return c.choice;
}
///////////////////////////////////////////////////
// RULESET EDITOR
///////////////////////////////////////////////////
int ruleseteditor(struct gamesettings* se){
	double upshiftaslevel=getspeedlv(se->initialupshifttime);
	char* stargennames[STARGEN_MAX] = {"Ends","Ends+Middle","Random"};
	struct cmenu c;
	char* labels[] = {
		"Number of Colors",
		"Board Width",
		"Board Height",
		"Initial Level",
		"Minimum Cores",
		"Maximum Cores",
		"Core Pattern",
		"Add Core Probability",
		"Minimum Block Length",
		"Maximum Block Length",
		"Minimum Rise Blocks",
		"Death Time",
		"Speed Increase Ratio",
		"Speed Increase Delay",
		"Block Place Time",
		"Rise Animation Time",
		"Single Tile Fall Time",
		"Block Clear Time",
		"Automatic Rise",
		"Number Initial Blocks",
		"Done",
	};
	c.numoptions=sizeof(labels)/sizeof(char*);
	c.labels=labels;
	void* values[] = {
		&se->numcolors,
		&se->defaultboardw,
		&se->defaultboardh,
		&upshiftaslevel,
		&se->minstars,
		&se->maxstars,
		stargennames[se->stargen],
		&se->staraddrarity,
		&se->minblocklen,
		&se->maxblocklen,
		&se->minimumupshiftblocksgenerated,
		&se->timetodie,
		&se->delayshortenratio,
		&se->delayshortentime,
		&se->playerplacetime,
		&se->upshiftanimtime,
		&se->falltime,
		&se->poptime,
		&se->doautoupshift,
		&se->initialblockgen,
		NULL,
	};
	char props[]={
		CMENUPROP_DISPLAYINT | CMENUPROP_LEFTRIGHT,
		CMENUPROP_DISPLAYINT | CMENUPROP_LEFTRIGHT,
		CMENUPROP_DISPLAYINT | CMENUPROP_LEFTRIGHT,
		CMENUPROP_DISPLAYDOUBLE | CMENUPROP_LEFTRIGHT | CMENUPROP_USERFLAG,
		CMENUPROP_DISPLAYINT | CMENUPROP_LEFTRIGHT,
		CMENUPROP_DISPLAYINT | CMENUPROP_LEFTRIGHT,
		CMENUPROP_LEFTRIGHT | CMENUPROP_USERFLAG,
		CMENUPROP_DISPLAYDOUBLE | CMENUPROP_LEFTRIGHT,
		CMENUPROP_DISPLAYINT | CMENUPROP_LEFTRIGHT,
		CMENUPROP_DISPLAYINT | CMENUPROP_LEFTRIGHT,
		CMENUPROP_DISPLAYINT | CMENUPROP_LEFTRIGHT,
		CMENUPROP_DISPLAYINT | CMENUPROP_LEFTRIGHT,
		CMENUPROP_DISPLAYDOUBLE | CMENUPROP_LEFTRIGHT,
		CMENUPROP_DISPLAYINT | CMENUPROP_LEFTRIGHT,
		CMENUPROP_DISPLAYINT | CMENUPROP_LEFTRIGHT,
		CMENUPROP_DISPLAYINT | CMENUPROP_LEFTRIGHT,
		CMENUPROP_DISPLAYINT | CMENUPROP_LEFTRIGHT,
		CMENUPROP_DISPLAYINT | CMENUPROP_LEFTRIGHT,
		CMENUPROP_DISPLAYBOOL | CMENUPROP_LEFTRIGHT,
		CMENUPROP_DISPLAYINT | CMENUPROP_LEFTRIGHT,
		0,
	};
	// min, max, increment
	double incrementamounts[]={
		1,maxcolors,1,
		2,(getScreenWidth()*.9)/blockw,1,
		2,getScreenHeight()/blockw,1,
		-99999,99999,-1,
		0,99,1,
		0,99,1,
		-1,-1,-1,
		0,1,.05,
		2,-1,1,
		2,-1,1,
		0,99,1,
		1,99999999,100,
		0,1,.05,
		1,99999999,1000,
		1,99999,20,
		1,999999,10,
		1,99999,10,
		1,99999,10,
		-1,-1,-1,
		0,99999,1,
		-1,-1,-1,
	};
	c.values=values;
	c.props=props;
	initcmenu(&c,getHDTime());
	char doupdatemins=1;
	char ret=1;
	u64 fTime;
	while(!cmenuoutrodone(&c, fTime=getHDTime())){
		controlsStart();
		int didselect=cmenucontrols(&c,fTime);
		if (!c.isoutro && (wasJustPressed(BUTTON_B) || wasJustPressed(BUTTON_START))){
			ret=!wasJustPressed(BUTTON_B);
			begincmenuoutro(&c, fTime);
		}
		controlsEnd();
		if (didselect){
			if (c.choice==c.numoptions-1){
				begincmenuoutro(&c, fTime);
				playSound(menuselectsound,0);
			}
			if (c.props[c.choice] & CMENUPROP_USERFLAG){
				if (c.choice==3){
					se->initialupshifttime=limitNum(se->initialupshifttime-100*didselect,10,MINLEVELUPSHIFTTIME*1029);
					upshiftaslevel=getspeedlv(se->initialupshifttime);
				}else if (c.choice==6){
					se->stargen=wrapnum(se->stargen+didselect,0,STARGEN_MAX-1);
					values[6]=stargennames[se->stargen];
				}
			}else{
				double min=incrementamounts[c.choice*3];
				double max=incrementamounts[c.choice*3+1];
				double inc=incrementamounts[c.choice*3+2]*didselect;
				if (c.props[c.choice] & CMENUPROP_DISPLAYINT){
					if (c.choice==8 && isDown(BUTTON_L)){
						incrementamounts[c.choice*3]=1;
						min=1;
					}
					int* dest=c.values[c.choice];
					*dest=limitNum(*dest+inc,min,max);
				}else if (c.props[c.choice] & CMENUPROP_DISPLAYDOUBLE){
					double* dest=c.values[c.choice];
					*dest=limitNumd(*dest+inc,min,max);
				}else if (c.props[c.choice] & CMENUPROP_DISPLAYBOOL){
					char* dest=c.values[c.choice];
					*dest=!(*dest);
				}
				// ensure board doesnt overflow
				if (c.choice==1 || c.choice==2 || c.choice==8 || c.choice==9){
					int oldmax=se->maxblocklen;
					se->maxblocklen=intmax(se->maxblocklen,se->minblocklen);
					int draww,drawh;
					getboarddimbysize(se, se->defaultboardw, se->defaultboardh, &draww, &drawh);
					if (draww>getScreenWidth() || drawh>getScreenHeight()){
						int* dest=c.values[c.choice];
						*dest-=inc;
						if (c.choice!=9){
							se->maxblocklen=oldmax;
						}
						startDrawing();
						drawRectangle(0,0,getScreenWidth(),getScreenHeight(),255,0,0,255);
						endDrawing();
					}
				}
				
			}
			doupdatemins=1;
		}
		if (doupdatemins){
			doupdatemins=0;
			// max block len
			incrementamounts[8*3+1]=intmin(se->defaultboardw,se->defaultboardh);
			incrementamounts[9*3+1]=intmin(se->defaultboardw,se->defaultboardh);
			incrementamounts[9*3]=se->minblocklen;
			// max block shapes
			incrementamounts[5*3]=se->minstars;
			
			for (int i=0;i<c.numoptions;++i){
				if (c.props[i] & CMENUPROP_DISPLAYINT){
					if (limitNumW(c.values[i],incrementamounts[i*3],incrementamounts[i*3+1])){
						cmenuvaluechange(&c,i);
					}
				}else if (c.props[i] & CMENUPROP_DISPLAYDOUBLE){
					if (limitNumdW(c.values[i],incrementamounts[i*3],incrementamounts[i*3+1])){
						cmenuvaluechange(&c,i);
					}
				}
			}
			cmenuvaluechange(&c,c.choice);
		}
		startDrawing();
		drawbackground();
		drawcmenu(&c,fTime);
		endDrawing();
	}
	return ret;
}
///////////////////////////////////////////////////
// CONTROLS EDITOR
///////////////////////////////////////////////////
static void defaultcontrols(struct pcontrols* c){
	initboundbutdef(&c->cw,BUTTON_A);
	initboundbutdef(&c->cc,BUTTON_B);
	initboundbutdef(&c->up,BUTTON_UP);
	initboundbutdef(&c->down,BUTTON_DOWN);
	initboundbutdef(&c->left,BUTTON_LEFT);
	initboundbutdef(&c->right,BUTTON_RIGHT);
	initboundbutdef(&c->place,BUTTON_X);
	addboundbut(&c->place, BUTTON_Y);
	initboundbutdef(&c->raisestack,BUTTON_L);
}
static void writeboundbut(struct boundbutton* b, FILE* fp){
	WRITETHING(b->numbuts,fp);
	for (int i=0;i<b->numbuts;++i){
		WRITETHING(b->buts[i],fp);
	}
}
static void readboundbut(struct boundbutton* b, FILE* fp){
	READTHING(b->numbuts,fp);
	b->buts=malloc(sizeof(crossButton)*b->numbuts);
	for (int i=0;i<b->numbuts;++i){
		READTHING(b->buts[i],fp);
	}
}
void savecontrols(const char* fpath, struct pcontrols* c, mltime dasdelay, mltime dasrepeat){
	FILE* fp = fopen(fpath,"wb");
	if (!fp){
		sadflash();
		return;
	}
	fputc(0x01,fp);
	WRITETHING(dasdelay,fp);
	WRITETHING(dasrepeat,fp);
	writeboundbut(&c->cw,fp);
	writeboundbut(&c->cc,fp);
	writeboundbut(&c->up,fp);
	writeboundbut(&c->down,fp);
	writeboundbut(&c->left,fp);
	writeboundbut(&c->right,fp);
	writeboundbut(&c->place,fp);
	writeboundbut(&c->raisestack,fp);
	fclose(fp);
}
void loadcontrols(const char* fpath, struct pcontrols* c, mltime* dasdelay, mltime* dasrepeat){
	FILE* fp = fopen(fpath,"rb");
	if (!fp){
	retdefault:
		defaultcontrols(c);
		*dasdelay=DEFAULTDASDELAY;
		*dasrepeat=DEFAULTDASREPEAT;
		return;
	}
	if (fgetc(fp)!=1){
		fclose(fp);
		goto retdefault;
	}
	READTHING(*dasdelay,fp);
	READTHING(*dasrepeat,fp);
	readboundbut(&c->cw,fp);
	readboundbut(&c->cc,fp);
	readboundbut(&c->up,fp);
	readboundbut(&c->down,fp);
	readboundbut(&c->left,fp);
	readboundbut(&c->right,fp);
	readboundbut(&c->place,fp);
	readboundbut(&c->raisestack,fp);
	fclose(fp);
}
static void _controlseditorsetvalues(void** values, char* currentbinds, char** bindnames, int numbuttons){
	for (int i=0;i<numbuttons;++i){
		values[i]=bindnames[currentbinds[i]];
	}
}
static void _controlseditorcalccurrentbinds(char* currentbinds, int numbinds, int numbuttons, struct boundbutton** bindtargets, crossButton* buttonmap){
	memset(currentbinds,numbinds,numbuttons); // Set them all to "Nothing"
	for (int i=0;i<numbinds;++i){
		for (int j=0;j<bindtargets[i]->numbuts;++j){
			for (int k=0;k<numbuttons;++k){
				if (bindtargets[i]->buts[j]==buttonmap[k]){
					currentbinds[k]=i;
					break;
				}
			}
		}
	}
}
void controlseditor(struct pcontrols* se){
	char* bindnames[] = {
		"Clockwise",
		"Counter CW",
		"Move Up",
		"Move Down",
		"Move Left",
		"Move Right",
		"Place Block",
		"Raise Stack",
		"Nothing",
	};
	struct boundbutton* bindtargets[] = {
		&se->cw,
		&se->cc,
		&se->up,
		&se->down,
		&se->left,
		&se->right,
		&se->place,
		&se->raisestack
	};
	int numbinds=sizeof(bindnames)/sizeof(char*)-1;
	
	char* labels[] = {
		"Cross",
		"Circle",
		"Triangle",
		"Square",
		"L Button",
		"R Button",
		"D-pad Up",
		"D-pad Down",
		"D-pad Left",
		"D-pad Right",
		"DAS Delay",
		"DAS Repeat",
		"Reset",
		"Done",
	};
	int numbuttons=10;
	crossButton buttonmap[/*numbuttons*/] = { // used to map buttons from labels into actual buttons
		BUTTON_A,
		BUTTON_B,
		BUTTON_X,
		BUTTON_Y,
		BUTTON_L,
		BUTTON_R,
		BUTTON_UP,
		BUTTON_DOWN,
		BUTTON_LEFT,
		BUTTON_RIGHT,
	};
	#if GBPLAT == GB_VITA
	if (vitaselectbutton==SCE_CTRL_CIRCLE){
		buttonmap[0]=BUTTON_B;
		buttonmap[1]=BUTTON_A;
	}
	#endif
	// easy way to tell what a button is currently bound to
	char currentbinds[numbuttons];
	_controlseditorcalccurrentbinds(currentbinds, numbinds, numbuttons, bindtargets, buttonmap);

	struct cmenu c;
	c.numoptions=numbuttons+4;
	c.labels=labels;
	
	void* values[c.numoptions];
	values[numbuttons]=&preferreddasdelay;
	values[numbuttons+1]=&preferreddasrepeat;
	values[c.numoptions-2]=NULL;
	values[c.numoptions-1]=NULL;
	_controlseditorsetvalues(values, currentbinds, bindnames, numbuttons);
	char props[c.numoptions];
	memset(props,CMENUPROP_LEFTRIGHT,numbuttons);
	// das
	props[numbuttons]=CMENUPROP_DISPLAYINT | CMENUPROP_LEFTRIGHT;
	props[numbuttons+1]=CMENUPROP_DISPLAYINT | CMENUPROP_LEFTRIGHT;
	// done/reset
	props[c.numoptions-2]=0;
	props[c.numoptions-1]=0;
	
	c.values=values;
	c.props=props;
	initcmenu(&c,getHDTime());
	u64 fTime;
	while(!cmenuoutrodone(&c, fTime=getHDTime())){
		controlsStart();
		int didselect=cmenucontrols(&c,fTime);
		if (!c.isoutro){
			if (wasJustPressed(BUTTON_B) || wasJustPressed(BUTTON_START)){
				playSound(menuselectsound,0);
				begincmenuoutro(&c, fTime);
				didselect=0;
			}
		}
		controlsEnd();
		if (didselect){
			if (c.choice==numbuttons || c.choice==numbuttons+1){
				int* dest=(int*)c.values[c.choice];
				*dest=limitNum(*dest+didselect*5,0,999999);
				cmenuvaluechange(&c, c.choice);
			}else{
				if (c.choice==c.numoptions-1){ // done
					playSound(menuselectsound,0);
					begincmenuoutro(&c, fTime);
				}else{
					if (c.choice==c.numoptions-2){
						playSound(diesound,0);
						defaultcontrols(se);
						preferreddasdelay=DEFAULTDASDELAY;
						preferreddasrepeat=DEFAULTDASREPEAT;
						_controlseditorcalccurrentbinds(currentbinds, numbinds, numbuttons, bindtargets, buttonmap);
						_controlseditorsetvalues(values, currentbinds, bindnames, numbuttons);
						cmenuvaluechangeall(&c);
					}else{
						currentbinds[c.choice]=wrapnum(currentbinds[c.choice]+didselect,0,numbinds);
						_controlseditorsetvalues(values, currentbinds, bindnames, numbuttons);
						cmenuvaluechange(&c, c.choice);
					}
				}
			}
		}
		startDrawing();
		drawbackground();
		drawcmenu(&c,fTime);
		endDrawing();
	}
	for (int i=0;i<numbinds;++i){
		freeboundbut(bindtargets[i]);
		initboundbut(bindtargets[i]);
	}
	for (int i=0;i<numbuttons;++i){
		if (currentbinds[i]<numbinds){
			addboundbut(bindtargets[currentbinds[i]], buttonmap[i]);
		}
	}
}
